#define BUTTON_BITMAP_H 16
#define BUTTON_BITMAP_W 16
extern unsigned char button_check_on_bits[];
extern unsigned char button_check_off_bits[];
extern unsigned char button_radio_on_bits[];
extern unsigned char button_radio_off_bits[];

#define MENU_BITMAP_H 11
#define MENU_BITMAP_W 11
extern unsigned char menu_check_on_bits[];
extern unsigned char menu_check_off_bits[];
extern unsigned char menu_radio_on_bits[];
extern unsigned char menu_radio_off_bits[];

#define MENU_ARROW_H 11
#define MENU_ARROW_W 11
extern unsigned char menu_arrow_bits[];

#define MENU_DOUBLE_ARROW_H 16
#define MENU_DOUBLE_ARROW_W 11
extern unsigned char menu_double_arrow_bits[];
