$LANGUAGES->{'de'} = {
                       '  The buttons in the navigation panels have the following meaning:' => '',
                       '  where the @strong{ Example } assumes that the current position is at @strong{ Subsubsection One-Two-Three } of a document of the following structure:' => '',
                       ' Up ' => '',
                       '%{acronym_like} (%{explanation})' => '',
                       '%{month}, %{day} %{year}' => '',
                       '%{name} of %{class}' => '',
                       '%{name} on %{class}' => '',
                       '%{node_file_href}' => '',
                       '%{node_file_href} @cite{%{book}}' => '',
                       '%{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       '%{reference_name}' => '',
                       '%{style} %{number}' => '',
                       '%{style}: %{caption_first_line}' => '',
                       '%{style}: %{shortcaption_first_line}' => '',
                       '@b{%{quotation_arg}:} ' => '',
                       '@cite{%{book}}' => '',
                       'About This Document' => '@"Uber dieses Dokument',
                       'April' => 'April',
                       'August' => 'August',
                       'Button' => '',
                       'Contents' => '',
                       'Current Position' => '',
                       'December' => 'Dezember',
                       'February' => 'Februar',
                       'Footnotes' => 'Fu@ss{}noten',
                       'From 1.2.3 go to' => '',
                       'Go to' => '',
                       'Index' => 'Index',
                       'Index Entry' => '',
                       'January' => 'Januar',
                       'July' => 'Juli',
                       'Jump to' => '',
                       'June' => 'Juni',
                       'March' => 'M@"arz',
                       'May' => 'Mai',
                       'Menu:' => '',
                       'Name' => '',
                       'Next' => '',
                       'November' => 'November',
                       'October' => 'Oktober',
                       'Overview' => '',
                       'Overview:' => '',
                       'Prev' => '',
                       'Section' => '',
                       'Section One' => '',
                       'See %{node_file_href}' => '',
                       'See %{node_file_href} @cite{%{book}}' => '',
                       'See %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'See %{reference_name}' => '',
                       'See @cite{%{book}}' => '',
                       'See section %{reference_name}' => '',
                       'See section `%{section}\' in @cite{%{book}}' => '',
                       'September' => 'September',
                       'Short Table of Contents' => 'Kurzes Inhaltsverzeichniss',
                       'Subsection One-Four' => '',
                       'Subsection One-One' => '',
                       'Subsection One-Three' => '',
                       'Subsection One-Two' => '',
                       'Subsubsection One-Two-Four' => '',
                       'Subsubsection One-Two-One' => '',
                       'Subsubsection One-Two-Three' => '',
                       'Subsubsection One-Two-Two' => '',
                       'T2H_today' => '',
                       'Table of Contents' => 'Inhaltsverzeichniss',
                       'This document was generated by @emph{%{user}} on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated by @emph{%{user}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated on @i{%{date}} using @uref{%{program_homepage}, @i{%{program}}}.' => '',
                       'This document was generated using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'Top' => '',
                       'Untitled Document' => '',
                       'about (help)' => '',
                       'beginning of this chapter or previous chapter' => '',
                       'by @emph{%{user}}' => '',
                       'by @emph{%{user}} on @emph{%{date}}' => '',
                       'cover (top) of document' => '',
                       'current' => '',
                       'current section' => '',
                       'first section in reading order' => '',
                       'following node' => '',
                       'index' => '',
                       'last section in reading order' => '',
                       'next chapter' => '',
                       'next node' => '',
                       'next section in reading order' => '',
                       'next section on same level' => '',
                       'node following in node reading order' => '',
                       'node up' => '',
                       'on @emph{%{date}}' => '',
                       'previous node' => '',
                       'previous section in reading order' => '',
                       'previous section on same level' => '',
                       'section `%{section}\' in @cite{%{book}}' => '',
                       'see %{node_file_href}' => '',
                       'see %{node_file_href} @cite{%{book}}' => '',
                       'see %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'see %{reference_name}' => '',
                       'see @cite{%{book}}' => '',
                       'see section %{reference_name}' => '',
                       'see section `%{section}\' in @cite{%{book}}' => '',
                       'short table of contents' => '',
                       'table of contents' => '',
                       'unknown' => '',
                       'up node' => '',
                       'up section' => ''
                     };

$T2H_OBSOLETE_STRINGS->{'de'} = {
                                  'See' => 'Siehe',
                                  'section' => 'Abschnitt',
                                  'see' => 'siehe'
                                };


$LANGUAGES->{'es'} = {
                       '  The buttons in the navigation panels have the following meaning:' => '',
                       '  where the @strong{ Example } assumes that the current position is at @strong{ Subsubsection One-Two-Three } of a document of the following structure:' => '',
                       ' Up ' => '',
                       '%{acronym_like} (%{explanation})' => '',
                       '%{month}, %{day} %{year}' => '',
                       '%{name} of %{class}' => '',
                       '%{name} on %{class}' => '',
                       '%{node_file_href}' => '',
                       '%{node_file_href} @cite{%{book}}' => '',
                       '%{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       '%{reference_name}' => '',
                       '%{style} %{number}' => '',
                       '%{style}: %{caption_first_line}' => '',
                       '%{style}: %{shortcaption_first_line}' => '',
                       '@b{%{quotation_arg}:} ' => '',
                       '@cite{%{book}}' => '',
                       'About This Document' => '',
                       'April' => 'abril',
                       'August' => 'agosto',
                       'Button' => '',
                       'Contents' => '',
                       'Current Position' => '',
                       'December' => 'diciembre',
                       'February' => 'febrero',
                       'Footnotes' => '',
                       'From 1.2.3 go to' => '',
                       'Go to' => '',
                       'Index' => 'Index',
                       'Index Entry' => '',
                       'January' => 'enero',
                       'July' => 'julio',
                       'Jump to' => '',
                       'June' => 'junio',
                       'March' => 'marzo',
                       'May' => 'mayo',
                       'Menu:' => '',
                       'Name' => '',
                       'Next' => '',
                       'November' => 'noviembre',
                       'October' => 'octubre',
                       'Overview' => '',
                       'Overview:' => '',
                       'Prev' => '',
                       'Section' => '',
                       'Section One' => '',
                       'See %{node_file_href}' => '',
                       'See %{node_file_href} @cite{%{book}}' => '',
                       'See %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'See %{reference_name}' => '',
                       'See @cite{%{book}}' => '',
                       'See section %{reference_name}' => '',
                       'See section `%{section}\' in @cite{%{book}}' => '',
                       'September' => 'septiembre',
                       'Short Table of Contents' => 'Resumen del Contenido',
                       'Subsection One-Four' => '',
                       'Subsection One-One' => '',
                       'Subsection One-Three' => '',
                       'Subsection One-Two' => '',
                       'Subsubsection One-Two-Four' => '',
                       'Subsubsection One-Two-One' => '',
                       'Subsubsection One-Two-Three' => '',
                       'Subsubsection One-Two-Two' => '',
                       'T2H_today' => '',
                       'Table of Contents' => '@\'{@dotless{I}}ndice General',
                       'This document was generated by @emph{%{user}} on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated by @emph{%{user}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated on @i{%{date}} using @uref{%{program_homepage}, @i{%{program}}}.' => '',
                       'This document was generated using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'Top' => '',
                       'Untitled Document' => '',
                       'about (help)' => '',
                       'beginning of this chapter or previous chapter' => '',
                       'by @emph{%{user}}' => '',
                       'by @emph{%{user}} on @emph{%{date}}' => '',
                       'cover (top) of document' => '',
                       'current' => '',
                       'current section' => '',
                       'first section in reading order' => '',
                       'following node' => '',
                       'index' => '',
                       'last section in reading order' => '',
                       'next chapter' => '',
                       'next node' => '',
                       'next section in reading order' => '',
                       'next section on same level' => '',
                       'node following in node reading order' => '',
                       'node up' => '',
                       'on @emph{%{date}}' => '',
                       'previous node' => '',
                       'previous section in reading order' => '',
                       'previous section on same level' => '',
                       'section `%{section}\' in @cite{%{book}}' => '',
                       'see %{node_file_href}' => '',
                       'see %{node_file_href} @cite{%{book}}' => '',
                       'see %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'see %{reference_name}' => '',
                       'see @cite{%{book}}' => '',
                       'see section %{reference_name}' => '',
                       'see section `%{section}\' in @cite{%{book}}' => '',
                       'short table of contents' => '',
                       'table of contents' => '',
                       'unknown' => '',
                       'up node' => '',
                       'up section' => ''
                     };

$T2H_OBSOLETE_STRINGS->{'es'} = {
                                  'See' => 'V@\'ease',
                                  'section' => 'secci@\'on',
                                  'see' => 'v@\'ase'
                                };


$LANGUAGES->{'fr'} = {
                       '  The buttons in the navigation panels have the following meaning:' => '  Les boutons de navigation ont la signification suivante :',
                       '  where the @strong{ Example } assumes that the current position is at @strong{ Subsubsection One-Two-Three } of a document of the following structure:' => '  Dans cet exemple on est @`a @strong{ Sous section un-deux-trois } dans un document dont la structure est :',
                       ' Up ' => 'Plus haut',
                       '%{acronym_like} (%{explanation})' => '',
                       '%{month}, %{day} %{year}' => 'le %{day} %{month} %{year}',
                       '%{name} of %{class}' => '%{name} de %{class}',
                       '%{name} on %{class}' => '%{name} de %{class}',
                       '%{node_file_href}' => '',
                       '%{node_file_href} @cite{%{book}}' => '',
                       '%{node_file_href} section `%{section}\' in @cite{%{book}}' => '%{node_file_href} section `%{section}\' dans @cite{%{book}}',
                       '%{reference_name}' => '',
                       '%{style} %{number}' => '',
                       '%{style}: %{caption_first_line}' => '',
                       '%{style}: %{shortcaption_first_line}' => '',
                       '@b{%{quotation_arg}:} ' => '',
                       '@cite{%{book}}' => '',
                       'About This Document' => 'A propos de ce document',
                       'April' => 'Avril',
                       'August' => 'Ao@^ut',
                       'Button' => 'Bouton',
                       'Contents' => 'Table des mati@`eres',
                       'Current Position' => 'Position',
                       'December' => 'D@\'ecembre',
                       'February' => 'F@\'evrier',
                       'Footnotes' => 'Notes de bas de page',
                       'From 1.2.3 go to' => 'Depuis 1.2.3 aller @`a',
                       'Go to' => 'Aller @`a',
                       'Index' => 'Index',
                       'Index Entry' => 'Entr@\'ee d\'index',
                       'January' => 'Janvier',
                       'July' => 'Juillet',
                       'Jump to' => 'Aller @`a',
                       'June' => 'Juin',
                       'March' => 'Mars',
                       'May' => 'Mai',
                       'Menu:' => 'Menu@ :',
                       'Name' => 'Nom',
                       'Next' => 'Suivant',
                       'November' => 'Novembre',
                       'October' => 'Octobre',
                       'Overview' => 'Vue d\'ensemble',
                       'Overview:' => 'Vue d\'ensemble@ :',
                       'Prev' => 'Pr@\'ec@\'edent',
                       'Section' => '',
                       'Section One' => 'Section un',
                       'See %{node_file_href}' => 'Voir %{node_file_href}',
                       'See %{node_file_href} @cite{%{book}}' => 'Voir %{node_file_href} @cite{%{book}}',
                       'See %{node_file_href} section `%{section}\' in @cite{%{book}}' => 'Voir %{node_file_href} section `%{section}\' dans @cite{%{book}}',
                       'See %{reference_name}' => 'Voir %{reference_name}',
                       'See @cite{%{book}}' => 'Voir @cite{%{book}}',
                       'See section %{reference_name}' => 'Voir la section %{reference_name}',
                       'See section `%{section}\' in @cite{%{book}}' => 'Voir la section `%{section}\' dans @cite{%{book}}',
                       'September' => 'Septembre',
                       'Short Table of Contents' => 'R@\'esum@\'e du contenu',
                       'Subsection One-Four' => 'Sous section un-quatre',
                       'Subsection One-One' => 'Sous section un-un',
                       'Subsection One-Three' => 'Sous section un-trois',
                       'Subsection One-Two' => 'Sous section un-deux',
                       'Subsubsection One-Two-Four' => 'Sous sous section un-deux-quatre',
                       'Subsubsection One-Two-One' => 'Sous sous section un-deux-un',
                       'Subsubsection One-Two-Three' => 'Sous sous section un-deux-trois',
                       'Subsubsection One-Two-Two' => 'Sous sous section un-deux-deux',
                       'T2H_today' => 'le %2$d %1$s %3$d',
                       'Table of Contents' => 'Table des mati@`eres',
                       'This document was generated by @emph{%{user}} on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e par @emph{%{user}} @emph{%{date}} en utilisant @uref{%{program_homepage}, @emph{%{program}}}.',
                       'This document was generated by @emph{%{user}} using @uref{%{program_homepage}, @emph{%{program}}}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e par @emph{%{user}} en utilisant @uref{%{program_homepage}, @emph{%{program}}}.',
                       'This document was generated on @i{%{date}} using @uref{%{program_homepage}, @i{%{program}}}.' => '',
                       'This document was generated using @uref{%{program_homepage}, @emph{%{program}}}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e en utilisant @uref{%{program_homepage}, @emph{%{program}}}.',
                       'Top' => '',
                       'Untitled Document' => 'Document sans titre',
                       'about (help)' => '@`a propos (page d\'aide)',
                       'beginning of this chapter or previous chapter' => 'd@\'ebut de ce chapitre ou chapitre pr@\'ec@\'edent',
                       'by @emph{%{user}}' => 'par @emph{%{user}}',
                       'by @emph{%{user}} on @emph{%{date}}' => 'par @emph{%{user}} @emph{%{date}}',
                       'cover (top) of document' => 'couverture (top) du document',
                       'current' => 'courante',
                       'current section' => 'section actuelle',
                       'first section in reading order' => 'premi@`e section dans l\'ordre de lecture',
                       'following node' => 'node suivant',
                       'index' => 'index',
                       'last section in reading order' => 'derni@`ere section dans l\'ordre de lecture',
                       'next chapter' => 'chapitre suivant',
                       'next node' => 'node suivant',
                       'next section in reading order' => 'section suivante dans l\'ordre de lecture',
                       'next section on same level' => 'section suivante au m@^eme niveau',
                       'node following in node reading order' => 'node suivant dans l\'ordre des nodes',
                       'node up' => 'node au dessus',
                       'on @emph{%{date}}' => '@emph{%{date}}',
                       'previous node' => 'node pr@\'ec@\'edent',
                       'previous section in reading order' => 'section pr@\'ec@\'edente dans l\'ordre de lecture',
                       'previous section on same level' => 'section pr@\'ec@\'edente au m@^eme niveau',
                       'section `%{section}\' in @cite{%{book}}' => 'section `%{section}\' dans @cite{%{book}}',
                       'see %{node_file_href}' => 'voir %{node_file_href}',
                       'see %{node_file_href} @cite{%{book}}' => 'voir %{node_file_href} @cite{%{book}}',
                       'see %{node_file_href} section `%{section}\' in @cite{%{book}}' => 'voir %{node_file_href} section `%{section}\' dans @cite{%{book}}',
                       'see %{reference_name}' => 'voir %{reference_name}',
                       'see @cite{%{book}}' => 'voir @cite{%{book}}',
                       'see section %{reference_name}' => 'voir la section %{reference_name}',
                       'see section `%{section}\' in @cite{%{book}}' => 'voir la section `%{section}\' dans @cite{{book}}',
                       'short table of contents' => 'table des mati@`eres r@\'esum@\'ee',
                       'table of contents' => 'table des mati@`eres',
                       'unknown' => 'inconnu',
                       'up node' => 'node au dessus',
                       'up section' => 'section sup@\'erieure'
                     };

$T2H_OBSOLETE_STRINGS->{'fr'} = {
                                  '  This document was generated %{who_and_when_generated} using %{program_homepage_href}.' => '  Ce document a &eacute;t&eacute; g&eacute;n&eacute;r&eacute; %{who_and_when_generated} en utilisant %{program_homepage_href}.',
                                  '  where the <strong> Example </strong> assumes that the current position is at <strong> Subsubsection One-Two-Three </strong> of a document of the following structure:' => '  Dans cet exemple on est &agrave; <strong> Sous section un-deux-trois </strong> dans un document dont la structure est :',
                                  '%{node_file_href} section `%{section}\' in <cite>%{book}</cite>' => '%{node_file_href} section `%{section}\' dans <cite>%{book}</cite>',
                                  'See' => 'Voir',
                                  'See %{node_file_href} <cite>%{book}</cite>' => 'Voir %{node_file_href} <cite>%{book}</cite>',
                                  'See %{node_file_href} section `%{section}\' in <cite>%{book}</cite>' => 'Voir %{node_file_href} section `%{section}\' dans <cite>%{book}</cite>',
                                  'See <cite>%{book}</cite>' => 'Voir <cite>%{book}</cite>',
                                  'See section `%{section}\' in <cite>%{book}</cite>' => 'Voir la section `%{section}\' dans <cite>%{book}</cite>',
                                  'This document was generated by <i>%{user}</i> on <i>%{date}</i> using %{program_homepage_href}.' => 'Ce document a &eacute;t&eacute; g&eacute;n&eacute;r&eacute; par <i>%{user}</i> <i>%{date}</i> en utilisant %{program_homepage_href}.',
                                  'This document was generated by <i>%{user}</i> using %{program_homepage_href}.' => 'Ce document a &eacute;t&eacute; g&eacute;n&eacute;r&eacute; par <i>%{user}</i> en utilisant %{program_homepage_href}.',
                                  'This document was generated by @emph{%{user}} on @emph{%{date}} using %{program_homepage_href}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e par @emph{%{user}} @emph{%{date}} en utilisant %{program_homepage_href}.',
                                  'This document was generated by @emph{%{user}} using %{program_homepage_href}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e par @emph{%{user}} en utilisant %{program_homepage_href}.',
                                  'This document was generated on <i>%{date}</i> using %{program_homepage_href}.' => 'Ce document a &eacute;t&eacute; g&eacute;n&eacute;r&eacute; <i>%{date}</i> en utilisant %{program_homepage_href}.',
                                  'This document was generated on @emph{%{date}} using %{program_homepage_href}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e @emph{%{date}} en utilisant %{program_homepage_href}.',
                                  'This document was generated on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e @emph{%{date}} en utilisant @uref{%{program_homepage}, @emph{%{program}}}.',
                                  'This document was generated using %{program_homepage_href}.' => 'Ce document a @\'et@\'e g@\'en@\'er@\'e en utilisant %{program_homepage_href}.',
                                  'about (this page)' => 'a propos (cette page)',
                                  'by <i>%{user}</i>' => 'par <i>%{user}</i>',
                                  'by <i>%{user}</i> on <i>%{date}</i>' => 'par <i>%{user}</i> <i>%{date}</i>',
                                  'concept index' => 'index',
                                  'on <i>%{date}</i>' => '<i>%{date}</i>',
                                  'section' => 'section',
                                  'section `%{section}\' in <cite>%{book}</cite>' => 'section `%{section}\' dans <cite>%{book}</cite>',
                                  'see' => 'voir',
                                  'see %{node_file_href} <cite>%{book}</cite>' => 'voir %{node_file_href} <cite>%{book}</cite>',
                                  'see %{node_file_href} section `%{section}\' in <cite>%{book}</cite>' => 'voir %{node_file_href} section `%{section}\' dans <cite>%{book}</cite>',
                                  'see <cite>%{book}</cite>' => 'voir <cite>%{book}</cite>',
                                  'see section `%{section}\' in <cite>%{book}</cite>' => 'voir la section `%{section}\' dans <cite>%{book}</cite>'
                                };


$LANGUAGES->{'nl'} = {
                       '  The buttons in the navigation panels have the following meaning:' => '',
                       '  where the @strong{ Example } assumes that the current position is at @strong{ Subsubsection One-Two-Three } of a document of the following structure:' => '',
                       ' Up ' => '',
                       '%{acronym_like} (%{explanation})' => '',
                       '%{month}, %{day} %{year}' => '',
                       '%{name} of %{class}' => '',
                       '%{name} on %{class}' => '',
                       '%{node_file_href}' => '',
                       '%{node_file_href} @cite{%{book}}' => '',
                       '%{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       '%{reference_name}' => '',
                       '%{style} %{number}' => '',
                       '%{style}: %{caption_first_line}' => '',
                       '%{style}: %{shortcaption_first_line}' => '',
                       '@b{%{quotation_arg}:} ' => '',
                       '@cite{%{book}}' => '',
                       'About This Document' => 'No translation available!',
                       'April' => 'April',
                       'August' => 'Augustus',
                       'Button' => '',
                       'Contents' => '',
                       'Current Position' => '',
                       'December' => 'December',
                       'February' => 'Februari',
                       'Footnotes' => 'No translation available!',
                       'From 1.2.3 go to' => '',
                       'Go to' => '',
                       'Index' => 'Index',
                       'Index Entry' => '',
                       'January' => 'Januari',
                       'July' => 'Juli',
                       'Jump to' => '',
                       'June' => 'Juni',
                       'March' => 'Maart',
                       'May' => 'Mei',
                       'Menu:' => '',
                       'Name' => '',
                       'Next' => '',
                       'November' => 'November',
                       'October' => 'Oktober',
                       'Overview' => '',
                       'Overview:' => '',
                       'Prev' => '',
                       'Section' => '',
                       'Section One' => '',
                       'See %{node_file_href}' => '',
                       'See %{node_file_href} @cite{%{book}}' => '',
                       'See %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'See %{reference_name}' => '',
                       'See @cite{%{book}}' => '',
                       'See section %{reference_name}' => '',
                       'See section `%{section}\' in @cite{%{book}}' => '',
                       'September' => 'September',
                       'Short Table of Contents' => 'Korte inhoudsopgave',
                       'Subsection One-Four' => '',
                       'Subsection One-One' => '',
                       'Subsection One-Three' => '',
                       'Subsection One-Two' => '',
                       'Subsubsection One-Two-Four' => '',
                       'Subsubsection One-Two-One' => '',
                       'Subsubsection One-Two-Three' => '',
                       'Subsubsection One-Two-Two' => '',
                       'T2H_today' => '',
                       'Table of Contents' => 'Inhoudsopgave',
                       'This document was generated by @emph{%{user}} on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated by @emph{%{user}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated on @i{%{date}} using @uref{%{program_homepage}, @i{%{program}}}.' => '',
                       'This document was generated using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'Top' => '',
                       'Untitled Document' => '',
                       'about (help)' => '',
                       'beginning of this chapter or previous chapter' => '',
                       'by @emph{%{user}}' => '',
                       'by @emph{%{user}} on @emph{%{date}}' => '',
                       'cover (top) of document' => '',
                       'current' => '',
                       'current section' => '',
                       'first section in reading order' => '',
                       'following node' => '',
                       'index' => '',
                       'last section in reading order' => '',
                       'next chapter' => '',
                       'next node' => '',
                       'next section in reading order' => '',
                       'next section on same level' => '',
                       'node following in node reading order' => '',
                       'node up' => '',
                       'on @emph{%{date}}' => '',
                       'previous node' => '',
                       'previous section in reading order' => '',
                       'previous section on same level' => '',
                       'section `%{section}\' in @cite{%{book}}' => '',
                       'see %{node_file_href}' => '',
                       'see %{node_file_href} @cite{%{book}}' => '',
                       'see %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'see %{reference_name}' => '',
                       'see @cite{%{book}}' => '',
                       'see section %{reference_name}' => '',
                       'see section `%{section}\' in @cite{%{book}}' => '',
                       'short table of contents' => '',
                       'table of contents' => '',
                       'unknown' => '',
                       'up node' => '',
                       'up section' => ''
                     };

$T2H_OBSOLETE_STRINGS->{'nl'} = {
                                  'See' => 'Zie',
                                  'section' => 'sectie',
                                  'see' => 'zie'
                                };


$LANGUAGES->{'no'} = {
                       '  The buttons in the navigation panels have the following meaning:' => '',
                       '  where the @strong{ Example } assumes that the current position is at @strong{ Subsubsection One-Two-Three } of a document of the following structure:' => '',
                       ' Up ' => '',
                       '%{acronym_like} (%{explanation})' => '',
                       '%{month}, %{day} %{year}' => '',
                       '%{name} of %{class}' => '',
                       '%{name} on %{class}' => '',
                       '%{node_file_href}' => '',
                       '%{node_file_href} @cite{%{book}}' => '',
                       '%{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       '%{reference_name}' => '',
                       '%{style} %{number}' => '',
                       '%{style}: %{caption_first_line}' => '',
                       '%{style}: %{shortcaption_first_line}' => '',
                       '@b{%{quotation_arg}:} ' => '',
                       '@cite{%{book}}' => '',
                       'About This Document' => 'No translation available!',
                       'April' => 'april',
                       'August' => 'august',
                       'Button' => '',
                       'Contents' => '',
                       'Current Position' => '',
                       'December' => 'desember',
                       'February' => 'februar',
                       'Footnotes' => 'No translation available!',
                       'From 1.2.3 go to' => '',
                       'Go to' => '',
                       'Index' => 'Indeks',
                       'Index Entry' => '',
                       'January' => 'januar',
                       'July' => 'juli',
                       'Jump to' => '',
                       'June' => 'juni',
                       'March' => 'mars',
                       'May' => 'mai',
                       'Menu:' => '',
                       'Name' => '',
                       'Next' => '',
                       'November' => 'november',
                       'October' => 'oktober',
                       'Overview' => '',
                       'Overview:' => '',
                       'Prev' => '',
                       'Section' => '',
                       'Section One' => '',
                       'See %{node_file_href}' => '',
                       'See %{node_file_href} @cite{%{book}}' => '',
                       'See %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'See %{reference_name}' => '',
                       'See @cite{%{book}}' => '',
                       'See section %{reference_name}' => '',
                       'See section `%{section}\' in @cite{%{book}}' => '',
                       'September' => 'september',
                       'Short Table of Contents' => 'Kort innholdsfortegnelse',
                       'Subsection One-Four' => '',
                       'Subsection One-One' => '',
                       'Subsection One-Three' => '',
                       'Subsection One-Two' => '',
                       'Subsubsection One-Two-Four' => '',
                       'Subsubsection One-Two-One' => '',
                       'Subsubsection One-Two-Three' => '',
                       'Subsubsection One-Two-Two' => '',
                       'T2H_today' => '',
                       'Table of Contents' => 'Innholdsfortegnelse',
                       'This document was generated by @emph{%{user}} on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated by @emph{%{user}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated on @i{%{date}} using @uref{%{program_homepage}, @i{%{program}}}.' => '',
                       'This document was generated using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'Top' => '',
                       'Untitled Document' => '',
                       'about (help)' => '',
                       'beginning of this chapter or previous chapter' => '',
                       'by @emph{%{user}}' => '',
                       'by @emph{%{user}} on @emph{%{date}}' => '',
                       'cover (top) of document' => '',
                       'current' => '',
                       'current section' => '',
                       'first section in reading order' => '',
                       'following node' => '',
                       'index' => '',
                       'last section in reading order' => '',
                       'next chapter' => '',
                       'next node' => '',
                       'next section in reading order' => '',
                       'next section on same level' => '',
                       'node following in node reading order' => '',
                       'node up' => '',
                       'on @emph{%{date}}' => '',
                       'previous node' => '',
                       'previous section in reading order' => '',
                       'previous section on same level' => '',
                       'section `%{section}\' in @cite{%{book}}' => '',
                       'see %{node_file_href}' => '',
                       'see %{node_file_href} @cite{%{book}}' => '',
                       'see %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'see %{reference_name}' => '',
                       'see @cite{%{book}}' => '',
                       'see section %{reference_name}' => '',
                       'see section `%{section}\' in @cite{%{book}}' => '',
                       'short table of contents' => '',
                       'table of contents' => '',
                       'unknown' => '',
                       'up node' => '',
                       'up section' => ''
                     };

$T2H_OBSOLETE_STRINGS->{'no'} = {
                                  'See' => 'Se',
                                  'section' => 'avsnitt',
                                  'see' => 'se'
                                };


$LANGUAGES->{'pt'} = {
                       '  The buttons in the navigation panels have the following meaning:' => '',
                       '  where the @strong{ Example } assumes that the current position is at @strong{ Subsubsection One-Two-Three } of a document of the following structure:' => '',
                       ' Up ' => '',
                       '%{acronym_like} (%{explanation})' => '',
                       '%{month}, %{day} %{year}' => '',
                       '%{name} of %{class}' => '',
                       '%{name} on %{class}' => '',
                       '%{node_file_href}' => '',
                       '%{node_file_href} @cite{%{book}}' => '',
                       '%{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       '%{reference_name}' => '',
                       '%{style} %{number}' => '',
                       '%{style}: %{caption_first_line}' => '',
                       '%{style}: %{shortcaption_first_line}' => '',
                       '@b{%{quotation_arg}:} ' => '',
                       '@cite{%{book}}' => '',
                       'About This Document' => '',
                       'April' => 'Abril',
                       'August' => 'Agosto',
                       'Button' => '',
                       'Contents' => '',
                       'Current Position' => '',
                       'December' => 'Dezembro',
                       'February' => 'Fevereiro',
                       'Footnotes' => '',
                       'From 1.2.3 go to' => '',
                       'Go to' => '',
                       'Index' => '@\'Indice',
                       'Index Entry' => '',
                       'January' => 'Janeiro',
                       'July' => 'Julho',
                       'Jump to' => '',
                       'June' => 'Junho',
                       'March' => 'Mar@,{c}o',
                       'May' => 'Maio',
                       'Menu:' => '',
                       'Name' => '',
                       'Next' => '',
                       'November' => 'Novembro',
                       'October' => 'Outubro',
                       'Overview' => '',
                       'Overview:' => '',
                       'Prev' => '',
                       'Section' => '',
                       'Section One' => '',
                       'See %{node_file_href}' => '',
                       'See %{node_file_href} @cite{%{book}}' => '',
                       'See %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'See %{reference_name}' => '',
                       'See @cite{%{book}}' => '',
                       'See section %{reference_name}' => '',
                       'See section `%{section}\' in @cite{%{book}}' => '',
                       'September' => 'Setembro',
                       'Short Table of Contents' => 'Breve Sum@\'ario',
                       'Subsection One-Four' => '',
                       'Subsection One-One' => '',
                       'Subsection One-Three' => '',
                       'Subsection One-Two' => '',
                       'Subsubsection One-Two-Four' => '',
                       'Subsubsection One-Two-One' => '',
                       'Subsubsection One-Two-Three' => '',
                       'Subsubsection One-Two-Two' => '',
                       'T2H_today' => '',
                       'Table of Contents' => 'Sum@\'ario',
                       'This document was generated by @emph{%{user}} on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated by @emph{%{user}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated on @i{%{date}} using @uref{%{program_homepage}, @i{%{program}}}.' => '',
                       'This document was generated using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'Top' => '',
                       'Untitled Document' => '',
                       'about (help)' => '',
                       'beginning of this chapter or previous chapter' => '',
                       'by @emph{%{user}}' => '',
                       'by @emph{%{user}} on @emph{%{date}}' => '',
                       'cover (top) of document' => '',
                       'current' => '',
                       'current section' => '',
                       'first section in reading order' => '',
                       'following node' => '',
                       'index' => '',
                       'last section in reading order' => '',
                       'next chapter' => '',
                       'next node' => '',
                       'next section in reading order' => '',
                       'next section on same level' => '',
                       'node following in node reading order' => '',
                       'node up' => '',
                       'on @emph{%{date}}' => '',
                       'previous node' => '',
                       'previous section in reading order' => '',
                       'previous section on same level' => '',
                       'section `%{section}\' in @cite{%{book}}' => '',
                       'see %{node_file_href}' => '',
                       'see %{node_file_href} @cite{%{book}}' => '',
                       'see %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'see %{reference_name}' => '',
                       'see @cite{%{book}}' => '',
                       'see section %{reference_name}' => '',
                       'see section `%{section}\' in @cite{%{book}}' => '',
                       'short table of contents' => '',
                       'table of contents' => '',
                       'unknown' => '',
                       'up node' => '',
                       'up section' => ''
                     };

$T2H_OBSOLETE_STRINGS->{'pt'} = {
                                  'See' => 'Veja',
                                  'section' => 'Se@,{c}@~ao',
                                  'see' => 'veja'
                                };


$LANGUAGES->{'en'} = {
                       '  The buttons in the navigation panels have the following meaning:' => '',
                       '  where the @strong{ Example } assumes that the current position is at @strong{ Subsubsection One-Two-Three } of a document of the following structure:' => '',
                       ' Up ' => '',
                       '%{acronym_like} (%{explanation})' => '',
                       '%{month}, %{day} %{year}' => '',
                       '%{name} of %{class}' => '',
                       '%{name} on %{class}' => '',
                       '%{node_file_href}' => '',
                       '%{node_file_href} @cite{%{book}}' => '',
                       '%{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       '%{reference_name}' => '',
                       '%{style} %{number}' => '',
                       '%{style}: %{caption_first_line}' => '',
                       '%{style}: %{shortcaption_first_line}' => '',
                       '@b{%{quotation_arg}:} ' => '',
                       '@cite{%{book}}' => '',
                       'About This Document' => '',
                       'April' => '',
                       'August' => '',
                       'Button' => '',
                       'Contents' => '',
                       'Current Position' => '',
                       'December' => '',
                       'February' => '',
                       'Footnotes' => '',
                       'From 1.2.3 go to' => '',
                       'Go to' => '',
                       'Index' => '',
                       'Index Entry' => '',
                       'January' => '',
                       'July' => '',
                       'Jump to' => '',
                       'June' => '',
                       'March' => '',
                       'May' => '',
                       'Menu:' => '',
                       'Name' => '',
                       'Next' => '',
                       'November' => '',
                       'October' => '',
                       'Overview' => '',
                       'Overview:' => '',
                       'Prev' => '',
                       'Section' => '',
                       'Section One' => '',
                       'See %{node_file_href}' => '',
                       'See %{node_file_href} @cite{%{book}}' => '',
                       'See %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'See %{reference_name}' => '',
                       'See @cite{%{book}}' => '',
                       'See section %{reference_name}' => '',
                       'See section `%{section}\' in @cite{%{book}}' => '',
                       'September' => '',
                       'Short Table of Contents' => '',
                       'Subsection One-Four' => '',
                       'Subsection One-One' => '',
                       'Subsection One-Three' => '',
                       'Subsection One-Two' => '',
                       'Subsubsection One-Two-Four' => '',
                       'Subsubsection One-Two-One' => '',
                       'Subsubsection One-Two-Three' => '',
                       'Subsubsection One-Two-Two' => '',
                       'T2H_today' => '%s, %d %d',
                       'Table of Contents' => '',
                       'This document was generated by @emph{%{user}} on @emph{%{date}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated by @emph{%{user}} using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'This document was generated on @i{%{date}} using @uref{%{program_homepage}, @i{%{program}}}.' => '',
                       'This document was generated using @uref{%{program_homepage}, @emph{%{program}}}.' => '',
                       'Top' => '',
                       'Untitled Document' => '',
                       'about (help)' => '',
                       'beginning of this chapter or previous chapter' => '',
                       'by @emph{%{user}}' => '',
                       'by @emph{%{user}} on @emph{%{date}}' => '',
                       'cover (top) of document' => '',
                       'current' => '',
                       'current section' => '',
                       'first section in reading order' => '',
                       'following node' => '',
                       'index' => '',
                       'last section in reading order' => '',
                       'next chapter' => '',
                       'next node' => '',
                       'next section in reading order' => '',
                       'next section on same level' => '',
                       'node following in node reading order' => '',
                       'node up' => '',
                       'on @emph{%{date}}' => '',
                       'previous node' => '',
                       'previous section in reading order' => '',
                       'previous section on same level' => '',
                       'section `%{section}\' in @cite{%{book}}' => '',
                       'see %{node_file_href}' => '',
                       'see %{node_file_href} @cite{%{book}}' => '',
                       'see %{node_file_href} section `%{section}\' in @cite{%{book}}' => '',
                       'see %{reference_name}' => '',
                       'see @cite{%{book}}' => '',
                       'see section %{reference_name}' => '',
                       'see section `%{section}\' in @cite{%{book}}' => '',
                       'short table of contents' => '',
                       'table of contents' => '',
                       'unknown' => '',
                       'up node' => '',
                       'up section' => ''
                     };

$T2H_OBSOLETE_STRINGS->{'en'} = {};


