#define VERSION 0

#define PACKAGE "package"

/* define if you have kpathsea */
#undef HAVE_KPATHSEA

/* define if your zlib has the compress2 function */
#undef HAVE_ZLIB_COMPRESS2

/* define if you have zlib and its headers */
#undef HAVE_ZLIB

/* define if you have xbasename() in your libkpathsea instead of basename() */
#undef HAVE_XBASENAME

/* define if kpathsea understands TrueType (TTF) formats */
#undef HAVE_TTF_FORMATS

/* define if struct tm has tm_gmtoff as a member */
#undef HAVE_TM_GMTOFF

/* define if <time.h> has timezone as an external variable */
#undef HAVE_TIMEZONE

/* define if kpathsea supports type1 and vf formats */
#undef HAVE_KPSE_FORMATS

/* define if you have libpng */
#undef HAVE_LIBPNG

/* Define if kpathsea understands Omega (OFM/OVF) formats */
#undef HAVE_OMEGA_FORMATS




