#define TFMPATH        "c:\\emtex\\tfm"
#define PKPATH         "c:\\texfonts\\pixel.lj\\%ddpi\\%f.%p"
#define VFPATH         "c:\\emtex\\vf"
#define HEADERPATH     ".;c:\\emtex\\ps"
#define CONFIGPATH     ".;c:\\emtex\\ps"
#define CONFIGFILE     "config.ps"
#define FONTSUBDIRPATH ""
#define FIGPATH        ".;..;c:\\emtex\\texinput"
#define FLIPATH        "c:\\texfonts"
#define FLINAME        "lj_0;lj_h;lj_1;lj_2;lj_3;lj_4;lj_5;lj_sli"
