/* regenerate c-auto.in from acconfig.h with:
   /usr/bin/autoheader -m ../etc/autoconf
   then run:
   /usr/bin/autoconf -m ../etc/autoconf
*/
/* c-auto.in.  Generated automatically from configure.in by autoheader.  */

/* Define if using alloca.c.  */
#undef C_ALLOCA

/* Define to one of _getb67, GETB67, getb67 for Cray-2 and Cray-YMP systems.
   This function is required for alloca.c support on those systems.  */
#undef CRAY_STACKSEG_END

/* Define if you have alloca, as a function or macro.  */
#undef HAVE_ALLOCA

/* Define if you have <alloca.h> and it should be used (not on Ultrix).  */
#undef HAVE_ALLOCA_H

/* Define if you don't have vprintf but do have _doprnt.  */
#undef HAVE_DOPRNT

/* Define if you have the ANSI # stringizing operator in cpp. */
#undef HAVE_STRINGIZE

/* Define if you have <sys/wait.h> that is POSIX.1 compatible.  */
#undef HAVE_SYS_WAIT_H

/* Define if you have <vfork.h>.  */
#undef HAVE_VFORK_H

/* Define if you have the vprintf function.  */
#undef HAVE_VPRINTF

/* Define to `int' if <sys/types.h> doesn't define.  */
#undef pid_t

/* Define as the return type of signal handlers (int or void).  */
#undef RETSIGTYPE

/* Define to `unsigned' if <sys/types.h> doesn't define.  */
#undef size_t

/* If using the C implementation of alloca, define if you know the
   direction of stack growth for your system; otherwise it will be
   automatically deduced at run-time.
 STACK_DIRECTION > 0 => grows toward higher addresses
 STACK_DIRECTION < 0 => grows toward lower addresses
 STACK_DIRECTION = 0 => direction of growth unknown
 */
#undef STACK_DIRECTION

/* Define if you have the ANSI C header files.  */
#undef STDC_HEADERS

/* Define vfork as fork if vfork does not work.  */
#undef vfork

/* Define if your processor stores words with the most significant
   byte first (like Motorola and SPARC, unlike Intel and VAX).  */
#undef WORDS_BIGENDIAN

/* Define if the X Window System is missing or not being used.  */
#undef X_DISPLAY_MISSING

/* Define if using alloca.c.  */
#undef C_ALLOCA

/* Define to one of _getb67, GETB67, getb67 for Cray-2 and Cray-YMP systems.
   This function is required for alloca.c support on those systems.  */
#undef CRAY_STACKSEG_END

/* Define if you have alloca, as a function or macro.  */
#undef HAVE_ALLOCA

/* Define if you have <alloca.h> and it should be used (not on Ultrix).  */
#undef HAVE_ALLOCA_H

/* Define if you don't have vprintf but do have _doprnt.  */
#undef HAVE_DOPRNT

/* Define if you have the ANSI # stringizing operator in cpp. */
#undef HAVE_STRINGIZE

/* Define if you have <sys/wait.h> that is POSIX.1 compatible.  */
#undef HAVE_SYS_WAIT_H

/* Define if you have <vfork.h>.  */
#undef HAVE_VFORK_H

/* Define if you have the vprintf function.  */
#undef HAVE_VPRINTF

/* Define to `int' if <sys/types.h> doesn't define.  */
#undef pid_t

/* Define as the return type of signal handlers (int or void).  */
#undef RETSIGTYPE

/* Define to `unsigned' if <sys/types.h> doesn't define.  */
#undef size_t

/* If using the C implementation of alloca, define if you know the
   direction of stack growth for your system; otherwise it will be
   automatically deduced at run-time.
 STACK_DIRECTION > 0 => grows toward higher addresses
 STACK_DIRECTION < 0 => grows toward lower addresses
 STACK_DIRECTION = 0 => direction of growth unknown
 */
#undef STACK_DIRECTION

/* Define if you have the ANSI C header files.  */
#undef STDC_HEADERS

/* Define vfork as fork if vfork does not work.  */
#undef vfork

/* Define if your processor stores words with the most significant
   byte first (like Motorola and SPARC, unlike Intel and VAX).  */
#undef WORDS_BIGENDIAN

/* Define if the X Window System is missing or not being used.  */
#undef X_DISPLAY_MISSING

/* c-auto.in.  Generated from configure.in by autoheader.  */
/* Define to determine the integer type to be used in bitmaps.  The type used
   will be "unsigned BMTYPE".  */
#define BMTYPE int

/* Define to the length (in bytes) of the above type.  */
#define BMBYTES 4

/* Define if your system has STREAMS (and if X uses it).  */
#undef HAVE_STREAMS

/* Define if you are using SunOS 4.x.  */
#undef SUNOS4

/* Define if you are using Linux 2.1.xxx -- 2.2.8,
   or if you find it necessary.  */
#undef FLAKY_SIGPOLL

/* Define if your system has <poll.h> and poll().  */
#undef HAVE_POLL

/* Define to use the Xaw toolkit.  */
#undef XAW

/* Define to use the Motif toolkit.  */
#undef MOTIF

/* Define to enable greyscale anti-aliasing for shrunken bitmaps.  */
#undef GREY

/* Define to enable support for color specials.  */
#undef COLOR

/* Define to enable gf format files (in addition to pk format).  */
#undef USE_GF

/* Define to use A4 as the default paper size.  */
#undef A4

/* Define to use hypertex */
#undef HTEX

/* Define to use ImageMagick */
#undef MAGICK

/* ??? */ 
#undef LDT1LIB

/* Define to compile in t1lib */
#undef T1LIB

/* Define to compile in t1lib */
#undef T1LIB

/* Define if your C compiler can do string concatenation */
#undef HAVE_CC_CONCAT

/* Define if the memicmp() function is in <string.h> */
#undef HAVE_MEMICMP

/* Define to point to the default command to use for printing (optional).  */
#undef DEFAULT_DVIPS_PATH

/* Define to point to the default command to use for printing (optional).  */
#undef DEFAULT_PS2PDF_PATH

/* Define if you have an old version of the Xaw library */
#undef HAVE_OLD_XAW

/* Define if you want to use the Xpm library */
#undef USE_XPM

/* Define to one of `_getb67', `GETB67', `getb67' for Cray-2 and Cray-YMP
   systems. This function is required for `alloca.c' support on those systems.
   */
#undef CRAY_STACKSEG_END

/* Define to 1 if using `alloca.c'. */
#undef C_ALLOCA

/* Define to 1 if you have `alloca', as a function or macro. */
#undef HAVE_ALLOCA

/* Define to 1 if you have <alloca.h> and it should be used (not on Ultrix).
   */
#undef HAVE_ALLOCA_H

/* Define to 1 if you don't have `vprintf' but do have `_doprnt.' */
#undef HAVE_DOPRNT

/* Define to 1 if cpp supports the ANSI # stringizing operator. */
#undef HAVE_STRINGIZE

/* Define to 1 if you have <sys/wait.h> that is POSIX.1 compatible. */
#undef HAVE_SYS_WAIT_H

/* Define to 1 if you have the <vfork.h> header file. */
#undef HAVE_VFORK_H

/* Define to 1 if you have the `vprintf' function. */
#undef HAVE_VPRINTF

/* Define as the return type of signal handlers (`int' or `void'). */
#undef RETSIGTYPE

/* If using the C implementation of alloca, define if you know the
   direction of stack growth for your system; otherwise it will be
   automatically deduced at run-time.
        STACK_DIRECTION > 0 => grows toward higher addresses
        STACK_DIRECTION < 0 => grows toward lower addresses
        STACK_DIRECTION = 0 => direction of growth unknown */
#undef STACK_DIRECTION

/* Define to 1 if you have the ANSI C header files. */
#undef STDC_HEADERS

/* Define to 1 if your processor stores words with the most significant byte
   first (like Motorola and SPARC, unlike Intel and VAX). */
#undef WORDS_BIGENDIAN

/* Define to 1 if the X Window System is missing or not being used. */
#undef X_DISPLAY_MISSING

/* Define to `int' if <sys/types.h> does not define. */
#undef pid_t

/* Define to `unsigned' if <sys/types.h> does not define. */
#undef size_t

/* Define as `fork' if `vfork' does not work. */
#undef vfork

/* Define to `unsigned long' if <sys/types.h> does not define. */
#undef uint32_t

/* Define if you have the fchdir function.  */
#undef HAVE_FCHDIR

/* Define if you have the ftruncate function.  */
#undef HAVE_FTRUNCATE

/* Define if you have the getcwd function.  */
#undef HAVE_GETCWD

/* Define if you have the getpwnam function.  */
#undef HAVE_GETPWNAM

/* Define if you have the getpwuid function.  */
#undef HAVE_GETPWUID

/* Define if you have the getuid function.  */
#undef HAVE_GETUID

/* Define if you have the isinf function.  */
#undef HAVE_ISINF

/* Define if you have the lstat function.  */
#undef HAVE_LSTAT

/* Define if you have the memcpy function.  */
#undef HAVE_MEMCPY

/* Define if you have the mkstemp function.  */
#undef HAVE_MKSTEMP

/* Define if you have the realpath function.  */
#undef HAVE_REALPATH

/* Define if you have the setenv function.  */
#undef HAVE_SETENV

/* Define if you have the sigaction function.  */
#undef HAVE_SIGACTION

/* Define if you have the snprintf function.  */
#undef HAVE_SNPRINTF

/* Define if you have the strchr function.  */
#undef HAVE_STRCHR

/* Define if you have the strerror function.  */
#undef HAVE_STRERROR

/* Define if you have the ulltostr function.  */
#undef HAVE_ULLTOSTR

/* Define if you have the vsnprintf function.  */
#undef HAVE_VSNPRINTF

/* Define if the vsnprintf function works.  */
#undef HAVE_GOOD_VSNPRINTF

/* Define if your system allows setsid() within vfork().  */
#undef HAVE_GOOD_SETSID_VFORK

/* Define if your system allows multiple definitions of functions.  */
#undef LD_ALLOWS_MULTIPLE_DEFINITIONS

/* Define if you have the waitpid function.  */
#undef HAVE_WAITPID

/* Define if you have the <X11/Xmu/Editres.h> header file.  */
#undef HAVE_X11_XMU_EDITRES_H

/* Define if you have the <X11/Xosdefs.h> header file.  */
#undef HAVE_X11_XOSDEFS_H

/* Define if you have the <X11/xpm.h> header file.  */
#undef HAVE_X11_XPM_H

/* Define if you have the <Xm/XpmP.h> header file.  */
#undef HAVE_XM_XPMP_H

/* Define if you have the <xpm.h> header file (not in X11, e.g. Solaris 5.8).  */
#undef HAVE_XPM_H

/* Define if you have the <iconv.h> header file.  */
#undef HAVE_ICONV_H

/* Define if you have the iconv function.  */
#undef HAVE_ICONV

/* Define the type of the iconv input string (char ** or const char **) */
#undef ICONV_CHAR_PPTR_TYPE

/* Define if you have the <ieeefp.h> header file.  */
#undef HAVE_IEEEFP_H

/* Define if you have the <netdb.h> header file.  */
#undef HAVE_NETDB_H

/* Define if you have the <regex.h> header file.  */
#undef HAVE_REGEX_H

/* Define if you have the <select.h> header file.  */
#undef HAVE_SELECT_H

/* Define if you have the <sys/bsdtypes.h> header file.  */
#undef HAVE_SYS_BSDTYPES_H

/* Define if you have the <sys/fcntl.h> header file.  */
#undef HAVE_SYS_FCNTL_H

/* Define if you have the <sys/param.h> header file.  */
#undef HAVE_SYS_PARAM_H

/* Define if you have the <sys/select.h> header file.  */
#undef HAVE_SYS_SELECT_H

/* Define if you have the <unistd.h> header file.  */
#undef HAVE_UNISTD_H

/* Define if the CODESET argument to nl_langinfo works.  */
#undef HAVE_WORKING_NL_LANGINFO_CODESET

/* Define to use Xaw panner. */
#undef USE_XAW_PANNER

/* Define to use libXp. */
#undef HAVE_LIB_XP

