(FAMILY UNSPECIFIED)
(FACE F MRR)
(CODINGSCHEME MP TEST: NOT A REAL FONT)
(DESIGNSIZE R 10.0)
(COMMENT DESIGNSIZE IS IN POINTS)
(COMMENT OTHER SIZES ARE MULTIPLES OF DESIGNSIZE)
(CHECKSUM O 0)
(FONTDIMEN
   (SLANT R -2.0)
   (SPACE R 0.400001)
   (STRETCH R 0.200001)
   (SHRINK R 0.1)
   (XHEIGHT R 0.45)
   (QUAD R 1.0)
   (EXTRASPACE R 0.200001)
   )
(CHARACTER O 0
   (CHARWD R 0.700001)
   (CHARHT R -0.1)
   (CHARDP R 0.200001)
   )
(CHARACTER O 55
   (CHARWD R -0.5)
   )
(CHARACTER O 57
   (CHARWD R 0.1)
   )
(CHARACTER C 1
   (CHARWD R 0.0)
   )
(CHARACTER C 2
   (CHARWD R 0.0)
   )
(CHARACTER C 3
   (CHARWD R 0.0)
   )
(CHARACTER C 4
   (CHARWD R 0.0)
   )
(CHARACTER C 5
   (CHARWD R 0.0)
   )
(CHARACTER C 6
   (CHARWD R 0.0)
   )
(CHARACTER C 7
   (CHARWD R 0.0)
   )
(CHARACTER C A
   (CHARWD R 0.200001)
   (CHARHT R 0.700001)
   (CHARDP R 0.1)
   (CHARIC R 0.1)
   )
(CHARACTER C B
   (CHARWD R 0.3)
   (CHARHT R 0.8)
   (CHARDP R 0.200001)
   )
(CHARACTER C C
   (CHARWD R 0.400001)
   )
(CHARACTER C M
   (CHARWD R 0.6)
   (CHARIC R 0.200001)
   (VARCHAR
      (MID C A)
      (BOT C B)
      (REP O 0)
      )
   )
(CHARACTER C a
   (CHARWD R 0.200001)
   (CHARHT R 0.700001)
   (CHARDP R 0.1)
   (CHARIC R 0.1)
   (NEXTLARGER O 202)
   )
(CHARACTER C b
   (CHARWD R 0.3)
   (CHARHT R 0.8)
   (CHARDP R 0.200001)
   (NEXTLARGER C M)
   )
(CHARACTER C p
   (CHARWD R 0.0)
   )
(CHARACTER C q
   (CHARWD R 0.0)
   )
(CHARACTER C r
   (CHARWD R 0.0)
   )
(CHARACTER C s
   (CHARWD R 0.0)
   )
(CHARACTER C t
   (CHARWD R 0.0)
   )
(CHARACTER C u
   (CHARWD R 0.0)
   )
(CHARACTER O 202
   (CHARWD R 0.400001)
   )
