//========================================================================
//
// FontEncodingTables.h
//
// Copyright 2001-2003 Glyph & Cog, LLC
//
//========================================================================

#ifndef FONTENCODINGTABLES_H
#define FONTENCODINGTABLES_H

extern char *macRomanEncoding[];
extern char *macExpertEncoding[];
extern char *winAnsiEncoding[];
extern char *standardEncoding[];
extern char *expertEncoding[];
extern char *symbolEncoding[];
extern char *zapfDingbatsEncoding[];

#endif
