/* ttf2afm.rc: ttf2afm				-*- C++ -*-

   Copyright (C) 1996-2003 Fabrice Popineau <fabrice.popineau@supelec.fr>

   This file is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2, or (at your
   option) any later version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file; if not, write to the Free Software
   Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. */

#include <texlive-version.h>

#define VER_FILEVERSION TEXLIVE_MAJOR_VERSION,TEXLIVE_MINOR_VERSION,TEXLIVE_BETA_NUMBER,TEXLIVE_ALPHA_NUMBER
#define VER_FILEVERSION_STR TEXLIVE_PRODUCTVERSION_STR " ttf2afm 1.10a"

#define VER_INTERNALNAME_STR "ttf2afm"
#define VER_ORIGINALFILENAME_STR "ttf2afm.exe"

#define VER_FILEDESCRIPTION_STR "TTF-to-AFM Converter"

#define VER_COMPANYNAME_STR         ""
#define VER_LEGALCOPYRIGHT_STR      "Copyright (c) 1996-2002 Han The Thanh, <thanh@pdftex.org>"

#include "texlive.version"

/* ttf2afm.rc ends here */
